"use strict";
/**
 * Parse ts code to type model
 * @desczh
 * 解析Typescript代码到type model
 * @file
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parse = exports.getModuleInfo = exports.getClasses = exports.parseInterface = exports.getintefaceChildren = exports.getType = exports.getTags = void 0;
var A = require("fp-ts/Array");
var O = require("fp-ts/Option");
var Ord_1 = require("fp-ts/Ord");
var path = require("path");
var pipeable_1 = require("fp-ts/pipeable");
var E = require("fp-ts/Either");
var ast = require("ts-morph");
var jsdoc_1 = require("./jsdoc");
var macoolka_object_1 = require("macoolka-object");
var macoolka_predicate_1 = require("macoolka-predicate");
var Monoid_1 = require("fp-ts/Monoid");
var macoolka_collection_1 = require("macoolka-collection");
var macoolka_type_model_1 = require("macoolka-type-model");
var macoolka_object_2 = require("macoolka-object");
var macoolka_store_core_1 = require("macoolka-store-core");
var Tags = {
    since: "since",
    file: "file",
    description: "description",
    examples: 'example',
    deprecated: 'deprecated',
    reason: 'reason',
    ignore: 'ignore',
    descriptions: 'descriptions'
};
/**
 * get comment info
 *
 * @since 0.2.0
 */
function getTags(jsdocs, isModule) {
    if (isModule === void 0) { isModule = false; }
    var jsdoc = (0, pipeable_1.pipe)(jsdocs.join('\n'), jsdoc_1.parseJSDoc);
    var descriptions = (0, pipeable_1.pipe)(jsdoc.tags, macoolka_collection_1.array.filter(function (a) { return a.title.startsWith('desc'); }), macoolka_collection_1.array.map(function (a) { return ({
        title: a.title.substring(4),
        desc: a.description
            ? a.description.split('\n')
            : []
    }); }), macoolka_collection_1.array.reduce({}, function (b, a) {
        var _a;
        return (__assign(__assign({}, b), (_a = {}, _a[a.title] = a.desc, _a)));
    }));
    var description = (0, pipeable_1.pipe)((0, macoolka_object_1.get)(jsdoc, ['description']), O.fromPredicate(macoolka_predicate_1.notEmpty), O.map(function (a) { return a.replace(/\\@/gi, '@'); }), O.map(function (a) { return a.split('\n'); }), O.getOrElse(function () { return []; }));
    return (0, pipeable_1.pipe)(jsdoc, jsdoc_1.tagToRecord, function (tag) { return ({
        since: (0, pipeable_1.pipe)((0, macoolka_object_1.get)(tag, [Tags.since, 'description']), O.fromPredicate(macoolka_predicate_1.notEmpty), O.toUndefined),
        description: description,
        examples: (0, pipeable_1.pipe)((0, macoolka_object_1.get)(tag, [Tags.examples, 'description']), O.fromPredicate(macoolka_predicate_1.notEmpty), O.map(function (a) { return a.split('\n'); }), O.getOrElse(function () { return []; })),
        file: (0, pipeable_1.pipe)((0, macoolka_object_1.get)(tag, [Tags.file]), function (a) { return a ? true : false; }),
        deprecated: (0, pipeable_1.pipe)((0, macoolka_object_1.get)(tag, [Tags.deprecated]), function (a) { return a ? true : false; }),
        ignore: (0, pipeable_1.pipe)((0, macoolka_object_1.get)(tag, [Tags.ignore]), function (a) { return a ? true : false; }),
        reason: (0, pipeable_1.pipe)((0, macoolka_object_1.get)(tag, [Tags.reason, 'description']), O.fromPredicate(macoolka_predicate_1.notEmpty), O.map(function (a) { return a.split('\n'); }), O.getOrElse(function () { return []; })),
        descriptions: descriptions,
    }); }, function (a) { return isModule ? a : (0, macoolka_object_2.omit)(a, 'file'); });
}
exports.getTags = getTags;
var getTypeText = function (a) { return stripImportTypes(a.getText().trim()); };
var getType = function (a, isArray) {
    if (isArray === void 0) { isArray = false; }
    var value = getTypeText(a);
    if (a.isBoolean()) {
        return {
            type: {
                _kind: 'boolean',
                isArray: isArray,
            }
        };
    }
    else if (a.isString()) {
        return {
            type: {
                _kind: 'string',
                isArray: isArray,
            }
        };
    }
    else if (a.isNumber()) {
        return {
            type: {
                _kind: 'number',
                isArray: isArray,
            }
        };
    }
    else if (a.isStringLiteral()) {
        return {
            type: {
                _kind: 'kind',
                value: value,
                isArray: isArray,
            }
        };
    }
    else if (a.isNumberLiteral()) {
        return {
            type: {
                _kind: 'kind',
                value: value,
                isArray: isArray,
            }
        };
    }
    else if (a.isBooleanLiteral()) {
        return {
            type: {
                _kind: 'kind',
                value: value,
                isArray: isArray,
            }
        };
    }
    else if (a.isArray()) {
        var type = a.getArrayElementType();
        var type1 = type ? (0, exports.getType)(type) : { type: { _kind: 'type', value: 'null' } };
        return type1;
    }
    else if (a.isClass()) {
        return {
            type: {
                _kind: 'type',
                value: value,
                isArray: isArray,
            }
        };
    }
    else if (a.isInterface()) {
        return {
            type: {
                _kind: 'type',
                value: value,
                isArray: isArray,
            }
        };
    }
    else if (a.isIntersection()) {
        var values = [];
        for (var _i = 0, _a = a.getIntersectionTypes(); _i < _a.length; _i++) {
            var g = _a[_i];
            values.push(getTypeText(g));
        }
        return {
            type: {
                _kind: 'typeUnion',
                values: values,
                isArray: isArray,
            }
        };
    }
    else if (a.isTuple()) {
        var values = [];
        for (var _b = 0, _c = a.getIntersectionTypes(); _b < _c.length; _b++) {
            var g = _c[_b];
            values.push(getTypeText(g));
        }
        return {
            type: {
                _kind: 'typeUnion',
                values: values,
                isArray: isArray,
            }
        };
    }
    else if (a.isUndefined()) {
        return {
            type: {
                _kind: 'type',
                value: 'undefined',
                isArray: isArray,
            }
        };
    }
    else if (a.isNull()) {
        return {
            type: {
                _kind: 'type',
                value: 'null',
                isArray: isArray,
            }
        };
    }
    else if (a.isUnion()) {
        var values = [];
        var literals = [];
        for (var _d = 0, _e = a.getUnionTypes(); _d < _e.length; _d++) {
            var g = _e[_d];
            if (g.isLiteral()) {
                literals.push(getTypeText(g));
            }
            else {
                values.push(getTypeText(g));
            }
        }
        if (values.length > 0) {
            return {
                type: {
                    _kind: 'typeUnion',
                    values: values,
                    isArray: isArray,
                }
            };
        }
        if (literals.length > 0) {
            return {
                type: {
                    _kind: 'enum',
                    values: literals,
                    isArray: isArray,
                }
            };
        }
        return {
            type: {
                _kind: 'typeUnion',
                values: values,
                isArray: isArray,
            }
        };
    }
    else if (a.isObject()) {
        return {
            type: {
                _kind: 'type',
                value: 'object',
                isArray: isArray,
            }
        };
    }
    else {
        return {
            type: {
                _kind: 'type',
                value: value,
                isArray: isArray,
            }
        };
    }
};
exports.getType = getType;
function getInteterfaceDeclarationSignature(id) {
    return id.getText();
}
var getName = function (a) {
    return (0, pipeable_1.pipe)(a, O.fromNullable, O.chain(function (a) { return (0, pipeable_1.pipe)(a.getName, O.fromPredicate(macoolka_predicate_1.isFunction), O.map(function (_) { return a.getName(); })); }));
};
function parseInterfaceDeclaration(id) {
    id.getProperties;
    var interf = __assign(__assign({ name: id.getName(), isExported: id.isExported(), signature: getInteterfaceDeclarationSignature(id), implements: (0, pipeable_1.pipe)(id.getBaseDeclarations(), A.map(function (a) { return getName(a); }), A.compact) }, getintefaceChildren(id)), (0, pipeable_1.pipe)(id.getJsDocs().map(function (a) { return a.getText(); }), getTags));
    return interf;
}
function getintefaceChildren(interD) {
    var methods = (0, pipeable_1.pipe)(interD.getMethods(), macoolka_collection_1.array.map(parseMethod));
    var _a = (0, pipeable_1.pipe)(interD.getProperties(), macoolka_collection_1.array.partition(function (vd) {
        // vd.getType().isAnonymous
        var initializer = vd.getInitializer();
        return (initializer && ast.Node.isFunctionLikeDeclaration(initializer)) ? true : false;
    }), function (_a) {
        var left = _a.left, right = _a.right;
        return ({
            left: left.map(parseProperty),
            right: right.map(parseMethodProperty)
        });
    }), left = _a.left, right = _a.right;
    var A = {
        methods: (0, pipeable_1.pipe)(methods.concat(right), macoolka_collection_1.array.sort(byName)),
        fields: (0, pipeable_1.pipe)(left, macoolka_collection_1.array.sort(byName))
    };
    return A;
}
exports.getintefaceChildren = getintefaceChildren;
var parseProperty = function (a) {
    var type = (0, exports.getType)(a.getType());
    a.hasQuestionToken;
    var field = __assign(__assign({ name: a.getName(), signature: a.getText(), required: !a.hasQuestionToken() }, type), (0, pipeable_1.pipe)(a.getJsDocs().map(function (a) { return a.getText(); }), getTags));
    return field;
};
function parseMethodProperty(a) {
    var type = (0, exports.getType)(a.getType());
    a.hasQuestionToken;
    var func = __assign(__assign({ name: a.getName(), signature: a.getText(), required: !a.hasQuestionToken() }, type), (0, pipeable_1.pipe)(a.getJsDocs().map(function (a) { return a.getText(); }), getTags));
    return func;
}
var parseMethod = function (a) {
    var type = (0, exports.getType)(a.getType());
    a.hasQuestionToken;
    var field = __assign(__assign({ name: a.getName(), signature: a.getText(), required: !a.hasQuestionToken() }, type), (0, pipeable_1.pipe)(a.getJsDocs().map(function (a) { return a.getText(); }), getTags));
    return field;
};
var byName = (0, pipeable_1.pipe)(Ord_1.ordString, (0, Ord_1.contramap)(function (x) { return x.name; }));
/**
 * @since 0.2.0
 */
function parseInterface(sourceFile) {
    return (0, pipeable_1.pipe)(sourceFile.getInterfaces(), macoolka_collection_1.array.map(parseInterfaceDeclaration), macoolka_collection_1.array.sort(byName));
}
exports.parseInterface = parseInterface;
function getFunctionDeclarationSignature(f) {
    var text = f.getText();
    return (0, pipeable_1.pipe)(f.compilerNode.body, O.fromNullable, O.map(function (body) {
        var end = body.getStart() - f.getStart() - 1;
        return text.substring(0, end) + ' { ... }';
    }), O.getOrElse(function () { return text + ' { ... }'; }));
}
function getFunctionDeclarationAnnotation(fd) {
    var overloads = fd.getOverloads();
    var docs = overloads.length === 0 ? fd.getJsDocs() : (overloads[0].getJsDocs());
    return docs.map(function (a) { return a.getText(); });
}
function parseFunctionDeclaration(fd) {
    var overloads = fd.getOverloads();
    var signatures = overloads.length === 0
        ? [getFunctionDeclarationSignature(fd)]
        : __spreadArray(__spreadArray([], overloads.slice(0, overloads.length - 1).map(function (fd) { return fd.getText(); }), true), [
            getFunctionDeclarationSignature(overloads[overloads.length - 1])
        ], false);
    var func = __assign({ name: fd.getName() ? fd.getName() : 'noName', signatures: signatures, isExported: fd.isExported() }, (0, pipeable_1.pipe)(getFunctionDeclarationAnnotation(fd), getTags));
    return func;
}
var indexOf = function (big, small) {
    var i = big.indexOf(small);
    return i !== -1 ? O.some(i) : O.none;
};
var lastIndexOf = function (big, small) {
    var i = big.lastIndexOf(small);
    return i !== -1 ? O.some(i) : O.none;
};
function getFunctionVariableDeclarationSignature(vd) {
    var text = vd.getText();
    var end = (0, pipeable_1.pipe)(indexOf(text, ' => {'), O.alt(function () { return lastIndexOf(text, ' =>'); }));
    return "export const ".concat(text.substring(0, (0, pipeable_1.pipe)(end, O.getOrElse(function () { return text.length; }))), " => ...");
}
function parseFunctionVariableDeclaration(vd) {
    var vs1 = vd.getParent();
    var vs = vs1.getParent();
    var func = __assign({ name: vd.getName(), signature: getFunctionVariableDeclarationSignature(vd), isExported: vd.isExported() }, (0, pipeable_1.pipe)(vs.getJsDocs().map(function (a) { return a.getText(); }), getTags));
    return func;
}
// fd.isExported() && !isInternal(getFunctionDeclarationAnnotation(fd))
function parseTypeAliasDeclaration(ta) {
    var typeAlias = __assign({ name: ta.getName(), signature: ta.getText() }, (0, pipeable_1.pipe)(ta.getJsDocs().map(function (a) { return a.getText(); }), getTags));
    return typeAlias;
}
/**
 * @since 0.2.0
 */
function getTypeAliases(sourceFile) {
    return (0, pipeable_1.pipe)(sourceFile.getTypeAliases(), macoolka_collection_1.array.map(parseTypeAliasDeclaration), macoolka_collection_1.array.sort(byName));
}
function getConstantVariableDeclarationSignature(vd) {
    var text = vd.getText();
    var end = text.indexOf(' = ');
    var name = text.substring(0, end);
    if (name.indexOf(':') === -1) {
        if (vd.getType && vd.getType().getText)
            name += ': ' + stripImportTypes(vd.getType().getText(vd));
        else
            name = 'empty';
    }
    return "export const ".concat(name, " = ...");
}
function parseConstantVariableDeclaration(vd) {
    var vs1 = vd.getParent();
    var vs = vs1.getParent();
    var constant = __assign({ name: vd.getName(), signature: getConstantVariableDeclarationSignature(vd), isExported: vd.isExported() }, (0, pipeable_1.pipe)(vs.getJsDocs().map(function (a) { return a.getText(); }), getTags));
    return constant;
}
/**
 * @since 0.2.0
 */
function getExport(sourceFile) {
    var methods = (0, pipeable_1.pipe)(sourceFile.getFunctions(), macoolka_collection_1.array.map(parseFunctionDeclaration));
    var _a = (0, pipeable_1.pipe)(sourceFile.getVariableDeclarations(), macoolka_collection_1.array.filter(function (a) { return a.isExported(); }), macoolka_collection_1.array.partition(function (vd) {
        var initializer = vd.getInitializer();
        return (initializer && ast.Node.isFunctionLikeDeclaration(initializer)) ? true : false;
    }), function (_a) {
        var left = _a.left, right = _a.right;
        return ({
            left: left.map(parseConstantVariableDeclaration),
            right: right.map(parseFunctionVariableDeclaration)
        });
    }), left = _a.left, right = _a.right;
    var A = {
        functions: (0, pipeable_1.pipe)(methods.concat(right), macoolka_collection_1.array.sort(byName)),
        constants: (0, pipeable_1.pipe)(left, macoolka_collection_1.array.sort(byName))
    };
    return A;
}
/**
 * @internal
 */
function stripImportTypes(s) {
    return s.replace(/import\("((?!").)*"\)./g, '');
}
function parseExportSpecifier(es) {
    var signature = stripImportTypes(es.getType().getText(es));
    var exp = {
        name: es.compilerNode.name.text,
        signature: signature,
    };
    return (0, pipeable_1.pipe)(A.head(es.getLeadingCommentRanges()), O.map(function (vs) { return (__assign(__assign({}, exp), (0, pipeable_1.pipe)([vs.getText()], getTags))); }), O.getOrElse(function () { return exp; }));
}
function parseExportDeclaration(ed) {
    return (0, pipeable_1.pipe)(ed.getNamedExports(), macoolka_collection_1.array.map(parseExportSpecifier));
}
/**
 * @since 0.2.0
 */
function getExports(sourceFile) {
    return (0, pipeable_1.pipe)(sourceFile.getExportDeclarations(), macoolka_collection_1.array.map(parseExportDeclaration), macoolka_collection_1.array.flatten, macoolka_collection_1.array.sort(byName));
}
function getTypeParameters(typeParameters) {
    return typeParameters.length === 0 ? '' : '<' + typeParameters.map(function (p) { return p.getName(); }).join(', ') + '>';
}
function getConstructorDeclarationSignature(c) {
    var text = c.getText();
    var body = c.compilerNode.body;
    if (body === undefined) {
        return text + ' { ... }';
    }
    var end = body.getStart() - c.getStart() - 1;
    return text.substring(0, end) + ' { ... }';
}
function getClassDeclarationSignature(c) {
    var dataName = c.getName();
    var typeParameters = getTypeParameters(c.getTypeParameters());
    var constructors = c.getConstructors();
    if (constructors.length > 0) {
        var constructorSignature = getConstructorDeclarationSignature(constructors[0]);
        return "export class ".concat(dataName).concat(typeParameters, " {\n  ").concat(constructorSignature, "\n  ... \n}");
    }
    else {
        return "export class ".concat(dataName).concat(typeParameters, " { ... }");
    }
}
function getMethodSignature(md) {
    var text = md.getText();
    var body = md.compilerNode.body;
    if (body === undefined) {
        return text + ' { ... }';
    }
    var end = body.getStart() - md.getStart() - 1;
    return text.substring(0, end) + ' { ... }';
}
function parseMethodDeclaration(md) {
    var overloads = md.getOverloads();
    var annotation = overloads.length === 0 ? md.getJsDocs() : overloads[0].getJsDocs();
    var signatures = overloads.length === 0
        ? [getMethodSignature(md)]
        : __spreadArray(__spreadArray([], overloads.slice(0, overloads.length - 1).map(function (md) { return md.getText(); }), true), [
            getMethodSignature(overloads[overloads.length - 1])
        ], false);
    var method = __assign({ name: md.getName(), signature: signatures.join('\n') }, (0, pipeable_1.pipe)(annotation.map(function (a) { return a.getText(); }), getTags));
    return method;
}
function parsePropertyDeclaration(md) {
    var annotation = md.getJsDocs();
    var signatures = md.getText();
    var method = __assign({ name: md.getName(), signature: signatures }, (0, pipeable_1.pipe)(annotation.map(function (a) { return a.getText(); }), getTags));
    return method;
}
function parseClass(c) {
    var name = c.getName();
    c.getProperties;
    var cls = __assign({ name: name ? name : 'noName', signature: getClassDeclarationSignature(c), fields: (0, pipeable_1.pipe)(c.getProperties(), macoolka_collection_1.array.map(parsePropertyDeclaration)), methods: (0, pipeable_1.pipe)(c.getInstanceMethods(), macoolka_collection_1.array.map(parseMethodDeclaration), macoolka_collection_1.array.sort(byName)), staticMethods: (0, pipeable_1.pipe)(c.getStaticMethods(), macoolka_collection_1.array.map(parseMethodDeclaration), macoolka_collection_1.array.sort(byName)) }, (0, pipeable_1.pipe)(c.getJsDocs().map(function (a) { return a.getText(); }), getTags));
    return cls;
}
/**
 * @since 0.2.0
 */
function getClasses(sourceFile) {
    return (0, pipeable_1.pipe)(sourceFile.getClasses(), macoolka_collection_1.array.map(parseClass), macoolka_collection_1.array.sort(byName));
}
exports.getClasses = getClasses;
/**
 * @since 0.2.0
 */
function getModuleInfo(sourceFile) {
    var x = sourceFile.getStatements();
    if (x.length > 0) {
        var comments = x[0].getLeadingCommentRanges();
        if (comments.length > 0) {
            return [comments[0].getText()];
        }
    }
    return undefined;
}
exports.getModuleInfo = getModuleInfo;
var ordModule = (0, pipeable_1.pipe)(Ord_1.ordString, (0, Ord_1.contramap)(function (module) { return (module.path ? module.path : []).join('/').toLowerCase(); }));
function getModuleName(p) {
    return path.parse(p[p.length - 1]).name;
}
function _parse(path, sourceFile) {
    var schema = __assign(__assign({ path: path, name: getModuleName(path), interfaces: parseInterface(sourceFile), typealiases: getTypeAliases(sourceFile), classes: getClasses(sourceFile), exports: getExports(sourceFile) }, getExport(sourceFile)), (0, pipeable_1.pipe)(getModuleInfo(sourceFile), function (a) { return a ? getTags(a, true) : {}; }));
    return schema;
}
/**
 * Parse Module from some typescript file path
 * @desczh
 * 解析typescript文件到‘typeModule’
 * @since 0.2.0
 */
function parse(files) {
    var paths = files.map(macoolka_store_core_1.showFile.show);
    var project = new ast.Project();
    paths.forEach(function (path) {
        project.addSourceFileAtPathIfExists(path);
    });
    return (0, pipeable_1.pipe)(paths, macoolka_collection_1.array.map(function (file) {
        var sourceFile = project.getSourceFile(file);
        return _parse(file.split(path.sep), sourceFile);
    }), macoolka_collection_1.array.sort(ordModule), macoolka_collection_1.array.map(function (a) { return (0, pipeable_1.pipe)(a, macoolka_type_model_1.io.NoValidNameModule.decode, macoolka_type_model_1.io.mapI18N); }), A.array.sequence(E.getValidation(Monoid_1.monoidString)));
}
exports.parse = parse;
exports.default = parse;
//# sourceMappingURL=parse.js.map